;;;; chicken-ffi.scm


(declare
 (disable-interrupts)
 (fixnum)
 (compress-literals 10) )

(library "chicken://ffi" "scheme://chicken"

  (export define-foreign-type define-foreign-variable foreign-lambda foreign-lambda*
	  foreign-safe-lambda foreign-safe-lambda* foreign-primitive foreign-safe-wrapper
	  define-external define-location let-location foreign-code foreign-value
	  define-foreign-record foreign-declare foreign-parse foreign-parse/declare)

  (import (for "chicken://chicken-internals" all))
  (import (for "scheme://chicken" all))
  (import (for "scheme://srfi-1" expand))
  (import (for "chicken://extras" expand))

  (begin-for-syntax
   (define (quotify xs)
     (map (lambda (x) #`',x) xs) ) )
    
  (define-syntax (define-foreign-type . xs)
    #`(##core#define-foreign-type
       ',(car xs)
       ',(cadr xs)
       ,@(cddr xs) ) )
  
  (define-syntax (define-foreign-variable . xs)
    #`(##core#define-foreign-variable ,@(quotify xs)) )

  (define-syntax (foreign-lambda . xs)
    #`(##core#foreign-lambda ,@(quotify xs)) )

  (define-syntax (foreign-lambda* . xs)
    #`(##core#foreign-lambda* ,@(quotify xs)) )

  (define-syntax (foreign-safe-lambda . xs)
    #`(##core#foreign-callback-lambda ,@(quotify xs)) )

  (define-syntax (foreign-safe-lambda* . xs)
    #`(##core#foreign-callback-lambda* ,@(quotify xs)) )

  (define-syntax (foreign-primitive . xs)
    #`(##core#foreign-primitive ,@(quotify xs)) )

  (define-syntax (foreign-safe-wrapper . form)
    (if (string? (caddr form))
	#`(##core#foreign-callback-wrapper
	   ',(cadr form)
	   ',(caddr form)
	   ',(car form)
	   ',(cadddr form)
	   ,(cadddr (cdr form)) )
	#`(##core#foreign-callback-wrapper
	   ',(cadr form)
	   '""
	   ',(car form)
	   ',(caddr form)
	   ,(cadddr form) ) ) )

  (define-syntax (define-external . form)
    (let* ([quals (and (pair? form) (string? (car form)))]
	   [var (and (not quals) (pair? form) (symbol? (car form)))] )
      (cond [var
	     (let ([var (car form)])
	       #`(begin
		  (##core#define-foreign-variable ',var ',(cadr form))
		  (##core#define-external-variable ',var ',(cadr form) '#t)
		  ,@(if (pair? (cddr form))
			#`((##core#set! ,var ,(caddr form)))
			#'() ) ) ) ]
	    [else
	     (let* ([head (if quals (cadr form) (car form))]
		    [args (cdr head)] )
	       #`(define ,(car head)
		  (##core#foreign-callback-wrapper
		   ',(car head)
		   ',(if quals (car form) "")
		   ',(if quals (caddr form) (cadr form))
		   ',(map (lambda (a) (car a)) args)
		   (lambda ,(map (lambda (a) (cadr a)) args) ,@(if quals (cdddr form) (cddr form)) ) ) ) ) ] ) ) )



;;; External locations:

  (define-syntax (define-location var type . init)
    (let ((name #`name))
      #`(begin
	 (##core#define-foreign-variable ',var ',type ',(symbol->string name))
	 (##core#define-external-variable ',var ',type '#f ',name)
	 ,@(if (pair? init)
	       #`((##core#set! ,var ,(car init)))
	       #'() ) ) ) ) 

  (define-syntax (let-location bindings . body)
    (let ([aliases (map (lambda _ #`tmp) bindings)])
      #`(let ,(append-map
	       (lambda (b a)
		 (if (pair? (cddr b))
		     (list (cons a (cddr b)))
		     #'() ) )
	       bindings aliases)
	  ,(fold-right
	    (lambda (b a rest)
	      (if (= 3 (length b))
		  #`(##core#let-location
		    (quote ,(car b))
		    (quote ,(cadr b))
		    ,a
		    ,rest)
		  #`(##core#let-location
		    (quote ,(car b))
		    (quote ,(cadr b))
		    ,rest) ) )
	    #`(let () ,@body)
	    bindings aliases) ) ) )


;;; Foreign records:

  (define-syntax (define-foreign-record name . slots)
    (let ([fname (if (pair? name) (cadr name) (sprintf "struct ~A" name))]
	  [tname (if (pair? name) (car name) name)] 
	  [var #`var]
	  [cvar #'cvar]
	  [xvar #`xvar]
	  [svar #`svar])
      (define (stype type)
	(cond [(not (pair? type)) type]
	      [(literal-identifier=? #'const (car type)) (stype (cadr type))]
	      [(or (literal-identifier=? (car type) #'struct) (literal-identifier=? (car type) #'union))
	       #`(pointer ,type)]
	      [else type] ) )
      #`(begin
	 ,@(if (pair? name)
	       #'() 
	       #`((declare
		   (foreign-declare
		    ,(string-intersperse
		      (append
		       (cons 
			(string-append "struct " (->string (syntax->datum name)) " { ")
			(map (lambda (slot)
			       (case (length slot)
				 [(3) 
				  (sprintf "~A[~A];"
					   (##compiler#foreign-type-declaration
					    (syntax->datum (car slot))
					    (->string (syntax->datum (cadr slot)) ) )
					   (caddr slot) ) ]
				 [(2)
				  (sprintf "~A;"
					   (##compiler#foreign-type-declaration 
					    (syntax->datum (car slot))
					    (->string (syntax->datum (cadr slot)) ) ) ) ]
				 [else (syntax-error "bad slot spec" slot)] ) )
			     slots) )
		       (list "};") )
		      "\n") ) ) ) )
	 (define-foreign-type ,tname (pointer ,fname))
	 ,@(map (lambda (slot)
		  (case (length slot)
		    [(3)
		     (let* ([type (syntax->datum (car slot))]
			    [sname (syntax->datum (cadr slot))]
			    [size (syntax->datum (caddr slot))]
			    [type2 (stype type)] )
		       #`(begin
			  (define ,(string->symbol (sprintf "~A-~A" tname sname))
			    (let ([,cvar 
				   (foreign-lambda* ,type2 ([,tname ,var] [int ,svar])
				     ,(sprintf "return(~A~A->~A[~A]);" 
					       (if (eq? type type2) "" "&")
					       var sname svar) ) ] )
			      (lambda (,var ,svar)
				(if (##core#check (and (fx>= ,svar 0) (fx< ,svar ,size)))
				    (,cvar ,var ,svar)
				    ;; this should signal a range exn...
				    (error "array access out of range" ',tname ',svar ,size) ) ) ) )
			  ,@(if (and (pair? type) (eq? 'const (car type))) 
				#'()
				(if (eq? type type2)
				    #`((define ,(string->symbol (sprintf "~A-~A-set!" tname sname))
					(let ([,cvar
					       (foreign-lambda* void ([,tname ,var] [int ,svar] [,type ,xvar])
						 ,(sprintf "~A->~A[~A] = ~A;" var sname svar xvar) ) ] )
					  (lambda (,var ,svar ,xvar)
					    (if (##core#check (and (fx>= ,svar 0) (fx< ,svar ,size)))
						(,cvar ,var ,svar ,xvar)
						(error "array access out of range" ',tname ',svar ,size) ) ) ) ) )
				    '() ) ) ) ) ]
		    [(2)
		     (let* ([type (syntax->datum (car slot))]
			    [sname (syntax->datum (cadr slot))] 
			    [type2 (stype type)] )  
		       #`(begin 
			  (define ,(string->symbol (sprintf "~A-~A" tname sname))
			    (foreign-lambda* ,type2 ([,tname ,var])
			      ,(sprintf "return(~A~A->~A);" 
					(if (eq? type type2) "" "&")
					var sname) ) )
			  ,@(if (and (pair? type) (eq? 'const (car type)))
				#'()
				(if (eq? type type2)
				    #`((define ,(string->symbol (sprintf "~A-~A-set!" tname sname))
					(foreign-lambda* void ([,tname ,var] [,type ,xvar])
					  ,(sprintf "~A->~A = ~A;" var sname xvar) ) ) )
				    #'() ) ) ) ) ]
		    [else (syntax-error "bad slot spec" slot)] ) )
		slots) ) ) )

  (define-syntax (foreign-declare . strs)
    #`(##core#declare '(foreign-declare ,@strs)))

  (define-syntax (foreign-parse . strs)
    #`(##core#declare '(foreign-parse ,@strs)))

  (define-syntax (foreign-parse/declare . strs)
    #`(##core#declare '(foreign-declare ,@strs) '(foreign-parse ,@strs)) )

  )
