;;;; chicken-macros-library.scm
;
; Note: no internal `define-values' and `eval-when' available


(declare
 (disable-interrupts)
 (fixnum)
 (compress-literals 10) )


(library "chicken://macros" "scheme://chicken"
  
  (export include define-macro receive time assert ensure :optional
	  critical-section nth-value define-record fluid-let
	  require-for-syntax parameterize set!-values and-let*
	  let-values let*-values define-inline define-constant
	  define-record-printer case-lambda letrec-values rec
	  condition-case define-record-type define-generic define-method
	  define-class cut cute let-string-start+end define-extension
	  switch let-optionals* handle-exceptions let-optionals)

  (import (for "chicken://chicken-internals" all))

  (define-syntax include
    (lambda (form)
      (if (= 2 (length form))
	  (let ((filename (cadr form)) )
	    #`(begin
		,@(datum->syntax-object 
		   (car form)
		   (with-input-from-file
		       (##sys#resolve-include-filename filename #t)
		     (lambda () (read-file))) ) ) )
	  (syntax-error "invalid `include' form") ) ) )

  (define-syntax define-macro
    (lambda (form)
      (cond ((and (= (length form) 3)
		  (identifier? (cadr form)))
	     (quasisyntax
	      (define-syntax ,(cadr form)
		(lambda (exp)
		  (datum->syntax-object
		   (car exp)
		   (apply ,(caddr form) (syntax-object->datum (cdr exp))))))))
	    ((and (>= (length form) 3)
		  (pair? (cadr form))
		  (identifier? (caadr form)))
	     (quasisyntax
	      (define-macro ,(caadr form)
		(lambda ,(cdadr form)
		  ,@(cddr form)))))
	    (else
	     (syntax-error "Define-macro:  Invalid syntax:" 
			   (syntax-debug form))))))
     
  (define-syntax (receive vars . more)
    (if (null? more)
	#`(##sys#call-with-values (lambda () ,vars) ##sys#list)
	#`(##sys#call-with-values
	   (lambda () ,(car more))
	   (lambda ,vars ,@(cdr more)) ) ) )

  (define-syntax (define-record name . ids)
    (define construct-name
      (lambda (template-identifier prefix . args)
	(datum->syntax-object
	 template-identifier
	 (##sys#string->qualified-symbol 
	  prefix
	  (##sys#apply string-append
		       (map (lambda (x)
			      (if (string? x)
				  x
				  (symbol->string (syntax-object->datum x))))
			    args))))))
    (let* ((nm (syntax-object->datum #'name))
	   (prefix (##sys#qualified-symbol-prefix nm))
	   (constructor (construct-name name prefix "make-" name))
	   (predicate (construct-name name prefix name "?"))
	   (access (map (lambda (x) (construct-name x prefix name "-" x)) ids))
	   (assign (map (lambda (x) (construct-name x prefix name "-" x "-set!")) ids))
	   (index
	    (let f ((i 1) (ids ids))
	      (if (null? ids)
		  '()
		  (cons i (f (+ i 1) (cdr ids)))))))
      #`(begin
	  (define ,constructor
	    (lambda ,ids
	      (##sys#make-structure ',name ,@ids)))
	  (define ,predicate
	    (lambda (x) (##sys#structure? x ',name)) )
	  ,@(map (lambda (a i)
		   #`(define ,a
		       (lambda (x)
			 (##sys#check-structure x ',name)
			 (##sys#block-ref x ,i) ) ) )
		 access index)
	  ,@(map (lambda (a i)
		   #`(define ,a
		       (lambda (x update)
			 (##sys#check-structure x ',name)
			 (##sys#block-set! x ,i update))) )
		 assign index) ) ) )

  (begin-for-syntax 
    
    (define (gentemps lst)
      (map (lambda _ #`tmp) lst) )
    
    ) ; begin-for-syntax

  (define-syntax (fluid-let defs . body)
    (let ((new (gentemps defs))
	  (vars (map car defs))
	  (vals (map cadr defs))
	  (old (gentemps defs)) )
      #`(let (,@(map (lambda (n v) #`(,n ,v)) new vals)
	      ,@(map (lambda (o) #`(,o #f)) old) )
	  (##sys#dynamic-wind
	   (lambda ()
	     ,@(map (lambda (o v) #`(##core#set! ,o ,v)) old vars)
	     ,@(map (lambda (n v) #`(##core#set! ,v ,n)) new vars) )
	   (lambda () ,@body)
	   (lambda ()
	     ,@(map (lambda (n v) #`(##core#set! ,n ,v)) new vars)
	     ,@(map (lambda (o v) #`(##core#set! ,v ,o)) old vars) ) ) ) ) )

  (define-syntax (parameterize defs . body)
    (let ((ps (gentemps defs))
	  (ys (gentemps defs))
	  (vars (map car defs))
	  (vals (map cadr defs)) )
      #`(let (,@(map (lambda (p x) #`(,p ,x)) ps vars)
	      ,@(map (lambda (y v) #`(,y ,v)) ys vals) )
	  (let ((swap (lambda () 
			,@(map (lambda (p y)
				 #`(let ((tmp (,p))) 
				     (,p ,y)
				     (##core#set! ,y tmp)) )
			       ps ys) ) ) )
	    (##sys#dynamic-wind swap (lambda () ,@body) swap)))))

  (define-syntax (require-for-syntax . names)
    #`(##core#require-for-syntax ,@names) )

  (define-syntax (define-values vars exp)
    #`(begin
	,@(map (lambda (v) #`(define ,v (##sys#void))) vars)
	(set!-values ,vars ,exp) ) )

  (define-syntax (set!-values vars exp)
    (if (null? vars)
	#`(##sys#call-with-values 
	   (lambda () ,exp) 
	   (lambda () (##core#undefined)) ) 
	(let ((tmps (gentemps vars)))
	  #`(##sys#call-with-values
	     (lambda () ,exp)
	     (lambda ,tmps
	       ,@(map (lambda (v t) #`(##core#set! ,v ,t)) vars tmps) ) ) ) ) )

  (define-syntax (and-let* defs . body)
    (let rec ((defs defs))
      (cond ((null? defs) #`(begin ,@body))
	    ((pair? (car defs))
	     (if (null? (cdar defs))
		 #`(if ,(caar defs) (and-let* ,(cdr defs) ,@body) #f)
		 #`(let (,(car defs))
		     (if ,(caar defs) (and-let* ,(cdr defs) ,@body) #f) ) ) )
	    (else #`(if ,(car defs) (and-let* ,(cdr defs) ,@body) #f) ) ) ) )

  (define-syntax (let*-values defs . body)
    (if (null? defs)
	#`(let () ,@body)
	#`(##sys#call-with-values 
	   (lambda () ,(cadar defs))
	   (lambda ,(caar defs)
	     (let*-values ,(cdr defs) ,@body) ) ) ) )

  (define-syntax (let-values defs . body)
    (let loop ((defs defs)
	       (tdefs '()) 
	       (body #`(let () ,@body)))
      (if (null? defs)
	  #`(let ,tdefs ,body)
	  (let ((tmps (gentemps (caar defs))))
	    #`(##sys#call-with-values
	       (lambda () ,(cadar defs))
	       (lambda ,tmps
		 ,(loop (cdr defs)
			(append (map list (caar defs) tmps) tdefs)
			body) ) ) ) ) ) )

  (define-syntax (define-inline head . body)
    #`(define ,head ,@body) )

  (define-syntax (define-constant name val)
    #`(define ,name ,val) )

  (define-syntax (define-record-printer head . body)
    (if (pair? head)
	#`(##sys#register-record-printer ',(car head) (lambda ,(cdr head) ,@body))
	#`(##sys#register-record-printer ',head ,@body)))

  (begin-for-syntax 
    
    (define (count-proper lst)
      (if (pair? lst)
          (fx+ 1 (count-proper (cdr lst)))
          0) )
    
    ) ; begin-for-syntax

  (define-syntax (case-lambda . clauses)
    (let ((num #`tmp)
	  (args #`args) )
      #`(lambda ,args
	  (let ((,num (length ,args)))
	    ,(let loop ((cs clauses))
	       (if (null? cs)		 
		   #'(##core#check 
		      (##sys#error
		       (##core#immutable '"wrong number of arguments to `case-lambda'")))
		   (let* ((c (car cs))
			  (vars (car c)) )
		     (cond ((list? vars)
			    #`(if (eq? ,num ,(length vars))
				  (##sys#apply (lambda ,vars ,@(cdr c)) ,args)
				  ,(loop (cdr cs)) ) )
			   ((pair? vars)
			    #`(if (fx>= ,num ,(count-proper vars))
				  (##sys#apply (lambda ,vars ,@(cdr c)) ,args)
				  ,(loop (cdr cs)) ) )
			   (else #`(let ((,vars ,args)) ,@(cdr c))) ) ) ) ) ) ) ) )

  (define-syntax (letrec-values defs . body)
    (let loop ((defs defs))
      (if (null? defs)
	  (let loop2 ((bindings defs))
	    (if (null? bindings)
		#`(let () ,@body)
		(let* ((vars (caar bindings))
		       (tmps (gentemps vars)) )
		  #`(##sys#call-with-values
		     (lambda () ,(cadar bindings))
		     (lambda ,tmps
		       ,@(map (lambda (v t) #`(##core#set! ,v ,t)) vars tmps)
		       ,(loop2 (cdr bindings)) ) ) ) ) )
	  #`(let ,(map (lambda (v) #`(,v (##core#undefined))) (caar defs))
	      ,(loop (cdr defs)) ) ) ) )

  (define-syntax (switch v . clauses)
    (let ((var #`tmp))
      #`(let ((,var ,v))
	  ,(let loop ((cs clauses))
	     (cond ((null? cs) #'(##core#undefined))
		   ((literal-identifier=? (caar cs) #'else)
		    #`(begin ,@(cdar cs)) )
		   (else #`(if (eqv? ,var ,(caar cs))
			       (begin ,@(cdar cs))
			       ,(loop (cdr cs)) ) ) ) ) ) ) )

  (define-syntax (let-optionals* rest defs . body)
    (let loop ((defs defs))
      (cond ((null? defs) #`(let () ,@body))
	    ((pair? (car defs))
	     #`(let* ((tmp ,rest)
		      (,(caar defs) (if (null? tmp) ,(cadar defs) (cdr tmp)))
		      (,rest (if (null? tmp) '() (cdr tmp))) )
		 ,(loop (cdr defs)) ) )
	    (else #`(let ((,(caar defs) ,rest)) ,@body) ) ) ) )

  (define-syntax (handle-exceptions var hbody . body)
    #`((call-with-current-continuation
	(lambda (k)
	  (with-exception-handler 
	   (lambda (,var) (k (lambda () ,hbody)))
	   (lambda ()
	     (##sys#call-with-values 
	      (lambda () ,@body)
	      (lambda args (k (lambda () (##sys#apply ##sys#values args)))))))))))

  (define-syntax (let-optionals arg-list var/defs . body)
    (define (make-default-procs vars body-proc defaulter-names defs rename)
      (let recur ((vars (reverse vars))
		  (defaulter-names (reverse defaulter-names))
		  (defs (reverse defs))
		  (next-guy body-proc))
	(if (null? vars) 
	    #'()
	    (let ((vars (cdr vars)))
	      #`((,(car defaulter-names)
		 (lambda ,(reverse vars)
		   (,next-guy ,@(reverse vars) ,(car defs))))
		 . ,(recur vars
			   (cdr defaulter-names)
			   (cdr defs)
			   (car defaulter-names)))))))
    
    (define (make-if-tree vars defaulters body-proc rest rename)
      (let recur ((vars vars) (defaulters defaulters) (non-defaults '()))
	(if (null? vars)
	    #`(if (##core#check (null? ,rest))
		  (,body-proc . ,(reverse non-defaults))
		  (##sys#error (##core#immutable '"too many optional arguments") ,rest))
	    (let ((v (car vars)))
	      #`(if (null? ,rest)
		   (,(car defaulters) . ,(reverse non-defaults))
		   (let ((,v (car ,rest))
			 (,rest (cdr ,rest)))
		     ,(recur (cdr vars)
			     (cdr defaulters)
			     (cons v non-defaults))))))))
    (let* ((vars (map car var/defs))
	   (vars2 (gentemps vars))
	   (defs (map cadr var/defs))
	   (body-proc #`body)
	   (rest-var #`%rest)
	   (defaulter-names (gentemps vars))
	   (defaulters (make-default-procs vars2 body-proc
					   defaulter-names defs gensym))
	   (if-tree (make-if-tree vars2 defaulter-names body-proc
				  rest-var gensym)))
      #`(let* ((,rest-var ,arg-list)
	       (,body-proc (lambda ,vars . ,body))
	       . ,defaulters)
	  ,if-tree) ) )

  (define-syntax (condition-case exp . clauses)
    (let ([exvar #`exvar]
	  [kvar #`kvar] )
      #`(handle-exceptions ,exvar
	   (let ([,kvar (and (##sys#structure? ,exvar 'condition) (##sys#slot ,exvar 1))])
	     ,(let rec ((clauses clauses))
		(if (null? clauses)
		    #`(##sys#signal ,exvar)
		    (let* ([c (car clauses)]
			   [var (and (identifier? (car c)) (car c))]
			   [kinds (if var (cadr c) (car c))]
			   [body (if var (cddr c) (cdr c))] )
		      (if (null? kinds)
			  (if var
			      #`(let ([,var ,exvar]) ,@body)
			      #`(let () ,@body) )
			  #`(if (and ,kvar ,@(map (lambda (k) #`(memv ',k ,kvar)) kinds))
				,(if var
				     #`(let ([,var ,exvar]) ,@body)
				     #`(let () ,@body) )
				,(rec (cdr clauses)) ) ) ) ) ) )
	  ,exp) ) )

  (begin-for-syntax
    
    (define (imemq id vars)
      (and (pair? vars)
           (or (literal-identifier=? id (car vars))
               (imemq id (cdr vars)))))
    
    ) ; begin-for-syntax

  (define-syntax (define-record-type t conser pred . slots)
    (let ([vars (cdr conser)]
	  [slotnames (map car slots)] )
      #`(begin
	  (define ,t ',t)
	  (define ,conser
	    (##sys#make-structure 
	     ',t 
	     ,@(map (lambda (sname)
		      (if (imemq sname vars)
			  sname
			  #'(##sys#void) ) )
		    slotnames) ) )
	  (define (,pred x) (##sys#structure? x ',t))
	  ,@(let loop ([slots slots] [i 1])
	      (if (null? slots)
		  #'()
		  (let ([slot (car slots)])
		    #`((define (,(cadr slot) x)
			(##sys#check-structure x ',t)
			(##sys#block-ref x ,i) )
		      ,@(if (pair? (cddr slot))
			    #`((define (,(caddr slot) x y)
				(##sys#check-structure x ',t)
				(##sys#block-set! x ,i y)) )
			    #'() )


		      ,@(loop (cdr slots) (fx+ i 1)) ) ) ) ) ) ) )

  (define-syntax (cut . more)
    (let loop ([xs more] [vars '()] [vals '()] [rest #f])
      (if (null? xs)
	  (let ([rvars (reverse vars)]
		[rvals (reverse vals)] )
	    (if rest
		(let ([rv #`rv])
		  #`(lambda (,@rvars . ,rv)
		      (apply ,(car rvals) ,@(cdr rvals) ,rv) ) )
		#`(lambda ,rvars ((begin ,(car rvals)) ,@(cdr rvals)) ) ) )
	  (let ((x (car xs)))
	    (cond ((literal-identifier=? x #'<>)
		   (let ([v #`tmp])
		     (loop (cdr xs) (cons v vars) (cons v vals) #f) ) )
		  ((literal-identifier=? x #'<...>) (loop '() vars vals #t))
		  (else (loop (cdr xs) vars (cons (car xs) vals) #f)) ) ) ) ) )

  (define-syntax (cute . more)
    (let loop ([xs more] [vars '()] [bs '()] [vals '()] [rest #f])
      (if (null? xs)
	  (let ([rvars (reverse vars)]
		[rvals (reverse vals)] )
	    (if rest
		(let ([rv #`rv])
		  #`(let ,bs
		      (lambda (,@rvars . ,rv)
			(apply ,(car rvals) ,@(cdr rvals) ,rv) ) ) )
		#`(let ,bs
		    (lambda ,rvars (,(car rvals) ,@(cdr rvals)) ) ) ) )
	  (let ((x (car xs)))
	    (cond ((literal-identifier=? x #'<>)
		   (let ([v #`tmp])
		     (loop (cdr xs) (cons v vars) bs (cons v vals) #f) ) )
		  ((literal-identifier=? x #'<...>) (loop '() vars bs vals #t))
		  (else 
		   (let ([v #`tmp])
		     (loop (cdr xs) vars (cons (list v (car xs)) bs) (cons v vals) #f) ) ) ) ) ) ) )

  (define-syntax (let-string-start+end s-e-r proc s-exp args-exp . body)
    (if (pair? (cddr s-e-r))
	#`(receive (,(caddr s-e-r) ,(car s-e-r) ,(cadr s-e-r))
	     (string-parse-start+end ,proc ,s-exp ,args-exp)
	   ,@body)
	#`(receive ,s-e-r
	     (string-parse-final-start+end ,proc ,s-exp ,args-exp)
	   ,@body) ) )

  (define-syntax (define-extension name . clauses)
    (let loop ((s '()) (d '()) (cs clauses) (exports #f))
      (cond ((null? cs)
	     (let ((exps (if exports #`(declare (export ,@exports)) #'(begin))))
	       #`(cond-expand
		 (chicken-compile-shared ,exps ,@d)
		 ((not compiling) ,@d)
		 (else 
		  (declare (unit ,name))
		  ,exps
		  (provide ',name) 
		  ,@s) ) ) )
	    ((and (pair? cs) (pair? (car cs)))
	     (let ((t (caar cs))
		   (next (cdr cs)) )
	       (cond ((literal-identifier=? #'static t) (loop (cons #`(begin ,@(cdar cs)) s) d next exports))
		     ((literal-identifier=? #'dynamic t) (loop s (cons #`(begin ,@(cdar cs)) d) next exports))
		     ((literal-identifier=? #'export t) (loop s d next (append (or exports '()) (cdar cs))))
		     (else (syntax-error "invalid clause specifier")))))
	    (else (syntax-error "invalid clause syntax")) ) ) )

  (define-syntax (rec head . args)
    (if (pair? head)
	#`(letrec ((,(car head) (lambda ,(cdr head) ,@args))) ,(car head))
	#`(letrec ((,head ,@args)) ,head)))

  ) ; module

